@@ -1,3 +1,4 @@
+from __future__ import absolute_import
 import weakref
 import copy
 import logging
@@ -6,41 +7,47 @@ import os
 import uuid
 
 import numpy as np
-from numpy.testing import assert_equal, assert_raises
-from nose import with_setup
-from nose.plugins.attrib import attr
+from numpy.testing import assert_equal
+import pytest
 
 from brian2 import (Clock, Network, ms, us, second, BrianObject, defaultclock,
                     run, stop, NetworkOperation, network_operation,
-                    restore_initial_state, MagicError, Synapses,
+                    MagicError, Synapses,
                     NeuronGroup, StateMonitor, SpikeMonitor,
                     SpikeGeneratorGroup,
                     PopulationRateMonitor, MagicNetwork, magic_network,
                     PoissonGroup, Hz, collect, store, restore, BrianLogger,
                     start_scope, prefs, profiling_summary, Quantity, TimedArray)
 from brian2.core.network import schedule_propagation_offset, scheduling_summary
-from brian2.devices.device import (reinit_devices, Device, all_devices,
-                                   set_device, get_device, reset_device, device)
+from brian2.devices.device import (reinit_and_delete, Device, all_devices,
+                                   set_device, get_device, reset_device, device,
+                                   RuntimeDevice)
 from brian2.utils.logger import catch_logs
 from brian2.tests.utils import assert_allclose
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_incorrect_network_use():
     '''Test some wrong uses of `Network` and `MagicNetwork`'''
-    assert_raises(TypeError, lambda: Network(name='mynet',
-                                             anotherkwd='does not exist'))
-    assert_raises(TypeError, lambda: Network('not a BrianObject'))
+    with pytest.raises(TypeError):
+        Network(name='mynet', anotherkwd='does not exist')
+    with pytest.raises(TypeError):
+        Network('not a BrianObject')
     net = Network()
-    assert_raises(TypeError, lambda: net.add('not a BrianObject'))
-    assert_raises(ValueError, lambda: MagicNetwork())
+    with pytest.raises(TypeError):
+        net.add('not a BrianObject')
+    with pytest.raises(ValueError):
+        MagicNetwork()
     G = NeuronGroup(10, 'v:1')
     net.add(G)
-    assert_raises(TypeError, lambda: net.remove(object()))
-    assert_raises(MagicError, lambda: magic_network.add(G))
-    assert_raises(MagicError, lambda: magic_network.remove(G))
+    with pytest.raises(TypeError):
+        net.remove(object())
+    with pytest.raises(MagicError):
+        magic_network.add(G)
+    with pytest.raises(MagicError):
+        magic_network.remove(G)
+        
 
-
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_network_contains():
     '''
     Test `Network.__contains__`.
@@ -51,8 +58,7 @@ def test_network_contains():
     assert 'neurongroup' not in net
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_empty_network():
     # Check that an empty network functions correctly
     net = Network()
@@ -68,8 +74,7 @@ class Counter(BrianObject):
         self.count += 1
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_network_single_object():
     # Check that a network with a single object functions correctly
     x = Counter()
@@ -78,8 +83,7 @@ def test_network_single_object():
     assert_equal(x.count, 10)
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_network_two_objects():
     # Check that a network with two objects and the same clock function correctly
     x = Counter(order=5)
@@ -87,15 +91,12 @@ def test_network_two_objects():
     net = Network()
     net.add([x, [y]]) # check that a funky way of adding objects work correctly
     net.run(1*ms)
-    assert_equal(net.objects[0].order, 5)
-    assert_equal(net.objects[1].order, 6)
     assert_equal(len(net.objects), 2)
     assert_equal(x.count, 10)
     assert_equal(y.count, 10)
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_network_from_dict():
     # Check that a network from a dictionary works
     x = Counter()
@@ -120,8 +121,7 @@ class NameLister(BrianObject):
         NameLister.updates.append(self.name)
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_network_different_clocks():
     NameLister.updates[:] = []
     # Check that a network with two different clocks functions correctly
@@ -133,8 +133,7 @@ def test_network_different_clocks():
     assert updates == ('xxxxxxxxxxy'*100000)
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_network_different_when():
     # Check that a network with different when attributes functions correctly
     NameLister.updates[:] = []
@@ -144,8 +143,7 @@ def test_network_different_when():
     net.run(0.3*ms)
     assert_equal(''.join(NameLister.updates), 'xyxyxy')
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_network_default_schedule():
     net = Network()
     assert net.schedule == ['start', 'groups', 'thresholds', 'synapses', 'resets', 'end']
@@ -153,8 +151,7 @@ def test_network_default_schedule():
     prefs.core.network.default_schedule = list(reversed(['start', 'groups', 'thresholds', 'synapses', 'resets', 'end']))
     assert net.schedule == list(reversed(['start', 'groups', 'thresholds', 'synapses', 'resets', 'end']))
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_network_schedule_change():
     # Check that a changed schedule is taken into account correctly
     NameLister.updates[:] = []
@@ -168,7 +165,7 @@ def test_network_schedule_change():
     net.run(0.3*ms)
     assert_equal(''.join(NameLister.updates), 'yxyxyx')
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_network_before_after_schedule():
     # Test that before... and after... slot names can be used
     NameLister.updates[:] = []
@@ -179,7 +176,7 @@ def test_network_before_after_schedule():
     net.run(0.3*ms)
     assert_equal(''.join(NameLister.updates), 'yxyxyx')
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_network_custom_slots():
     # Check that custom slots can be inserted into the schedule
     NameLister.updates[:] = []
@@ -191,25 +188,30 @@ def test_network_custom_slots():
     net.run(0.3*ms)
     assert_equal(''.join(NameLister.updates), 'xyzxyzxyz')
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_network_incorrect_schedule():
     # Test that incorrect arguments provided to schedule raise errors
     net = Network()
     # net.schedule = object()
-    assert_raises(TypeError, setattr, net, 'schedule', object())
+    with pytest.raises(TypeError):
+        setattr(net, 'schedule', object())
     # net.schedule = 1
-    assert_raises(TypeError, setattr, net, 'schedule', 1)
+    with pytest.raises(TypeError):
+        setattr(net, 'schedule', 1)
     # net.schedule = {'slot1', 'slot2'}
-    assert_raises(TypeError, setattr, net, 'schedule', {'slot1', 'slot2'})
+    with pytest.raises(TypeError):
+        setattr(net, 'schedule', {'slot1', 'slot2'})
     # net.schedule = ['slot', 1]
-    assert_raises(TypeError, setattr, net, 'schedule', ['slot', 1])
+    with pytest.raises(TypeError):
+        setattr(net, 'schedule', ['slot', 1])
     # net.schedule = ['start', 'after_start']
-    assert_raises(ValueError, setattr, net, 'schedule', ['start', 'after_start'])
+    with pytest.raises(ValueError):
+        setattr(net, 'schedule', ['start', 'after_start'])
     # net.schedule = ['before_start', 'start']
-    assert_raises(ValueError, setattr, net, 'schedule', ['before_start', 'start'])
+    with pytest.raises(ValueError):
+        setattr(net, 'schedule', ['before_start', 'start'])
 
-@attr('codegen-independent')
-@with_setup(teardown=reinit_devices)
+@pytest.mark.codegen_independent
 def test_schedule_warning():
     previous_device = get_device()
     from uuid import uuid4
@@ -268,7 +270,7 @@ def test_schedule_warning():
     reset_device(previous_device)
 
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_scheduling_summary_magic():
     basename = 'name' + str(uuid.uuid4()).replace('-', '_')
     group = NeuronGroup(10, 'dv/dt = -v/(10*ms) : 1', threshold='v>1',
@@ -311,7 +313,7 @@ def test_scheduling_summary_magic():
     assert summary_after._repr_html_() == summary_before._repr_html_()
 
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_scheduling_summary():
     basename = 'name' + str(uuid.uuid4()).replace('-', '_')
     group = NeuronGroup(10, 'dv/dt = -v/(10*ms) : 1', threshold='v>1',
@@ -322,24 +324,30 @@ def test_scheduling_summary():
                                       name=basename + '_sm_ia',
                                       when='after_end')
     inactive_state_mon.active = False
-    net = Network(group, state_mon, inactive_state_mon)
+    @network_operation(name=basename + '_net_op', when='before_end')
+    def foo():
+        pass
+    net = Network(group, state_mon, inactive_state_mon, foo)
     summary_before = scheduling_summary(net)
     assert [entry.name for entry in summary_before.entries] == [basename+'_sm',
                                                                 basename+'_stateupdater',
                                                                 basename+'_thresholder',
                                                                 basename+'_resetter',
+                                                                basename+'_net_op',
                                                                 basename+'_run_regularly',
                                                                 basename+'_sm_ia']
     assert [entry.when for entry in summary_before.entries] == ['start',
                                                                 'groups',
                                                                 'thresholds',
                                                                 'resets',
+                                                                'before_end',
                                                                 'end',
                                                                 'after_end']
     assert [entry.dt for entry in summary_before.entries] == [defaultclock.dt,
                                                               defaultclock.dt,
                                                               defaultclock.dt,
                                                               defaultclock.dt,
+                                                              defaultclock.dt,
                                                               defaultclock.dt*10,
                                                               defaultclock.dt]
     assert [entry.active for entry in summary_before.entries] == [True,
@@ -347,6 +355,7 @@ def test_scheduling_summary():
                                                                   True,
                                                                   True,
                                                                   True,
+                                                                  True,
                                                                   False]
     assert len(str(summary_before))
     assert len(summary_before._repr_html_())
@@ -371,8 +380,7 @@ class Preparer(BrianObject):
         self.did_post_run = True        
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_magic_network():
     # test that magic network functions correctly
     x = Counter()
@@ -396,8 +404,7 @@ class Stopper(BrianObject):
         if self.stoptime<=0:
             self.stopfunc()
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_network_stop():
     # test that Network.stop and global stop() work correctly
     net = Network()
@@ -412,8 +419,7 @@ def test_network_stop():
     assert_equal(defaultclock.t, 1*ms)
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_network_operations():
     # test NetworkOperation and network_operation
     seq = []
@@ -448,7 +454,7 @@ def test_network_operations():
 
     assert_equal(''.join(seq), 'bBacC'*10)
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_incorrect_network_operations():
     # Network operations with more than one argument are not allowed
     def func(x, y):
@@ -459,8 +465,10 @@ def test_incorrect_network_operations():
             pass
     c = Container()
 
-    assert_raises(TypeError, lambda: NetworkOperation(func))
-    assert_raises(TypeError, lambda: NetworkOperation(c.func))
+    with pytest.raises(TypeError):
+        NetworkOperation(func)
+    with pytest.raises(TypeError):
+        NetworkOperation(c.func)
 
     # Incorrect use of @network_operation -- it does not work on an instance
     # method
@@ -473,8 +481,35 @@ def test_incorrect_network_operations():
     except TypeError:
         pass  # this is what we expected
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
+def test_network_operations_name():
+    # test NetworkOperation name input
+    seq = []
+    def f1():
+        seq.append('a')
+
+    def f2():
+        seq.append('b')
+
+    def x():
+        pass
+    op = NetworkOperation(lambda: x)
+    assert_equal(op.name, 'networkoperation')
+
+    op0 = NetworkOperation(lambda: x, name='named_network')
+    assert_equal(op0.name, 'named_network')
+
+    op1 = NetworkOperation(f1, name='networkoperation_1')
+    op2 = NetworkOperation(f1, name='networkoperation_3')
+    op3 = NetworkOperation(f2, name='networkoperation_2')
+
+    net = Network(op1, op2, op3)
+    net.run(1*ms)
+
+    assert_equal(''.join(seq), 'aba'*10)
+
+
+@pytest.mark.codegen_independent
 def test_network_active_flag():
     # test that the BrianObject.active flag is recognised by Network.run
     x = Counter()
@@ -485,8 +520,7 @@ def test_network_active_flag():
     assert_equal(y.count, 0)
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_network_t():
     # test that Network.t works as expected
     x = Counter(dt=1*ms)
@@ -530,30 +564,29 @@ def test_network_t():
     assert_equal(y.count, 5)
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_incorrect_dt_defaultclock():
     defaultclock.dt = 0.5*ms
     G = NeuronGroup(1, 'dv/dt = -v / (10*ms) : 1')
     net = Network(G)
     net.run(0.5*ms)
     defaultclock.dt = 1*ms
-    assert_raises(ValueError, lambda: net.run(0*ms))
+    with pytest.raises(ValueError):
+        net.run(0*ms)
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_incorrect_dt_custom_clock():
     clock = Clock(dt=0.5*ms)
     G = NeuronGroup(1, 'dv/dt = -v / (10*ms) : 1', clock=clock)
     net = Network(G)
     net.run(0.5*ms)
     clock.dt = 1*ms
-    assert_raises(ValueError, lambda: net.run(0*ms))
+    with pytest.raises(ValueError):
+        net.run(0*ms)
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_network_remove():
     x = Counter()
     y = Counter()
@@ -577,8 +610,7 @@ class NoninvalidatingCounter(Counter):
     add_to_magic_network = True
     invalidates_magic_network = False
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_invalid_magic_network():
     x = Counter()
     run(1*ms)
@@ -611,8 +643,7 @@ def test_invalid_magic_network():
     assert_equal(y.count, 10) 
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_multiple_networks_invalid():
     x = Counter()
     net = Network(x)
@@ -630,8 +661,7 @@ def test_multiple_networks_invalid():
         pass  # this is expected
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_magic_weak_reference():
     '''
     Test that holding a weak reference to an object does not make it get
@@ -650,8 +680,7 @@ def test_magic_weak_reference():
         assert '2 objects' in magic_objects, 'Unexpected log message: %s' % magic_objects
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_magic_unused_object():
     '''Test that creating unused objects does not affect the magic system.'''
     def create_group():
@@ -670,8 +699,7 @@ def test_magic_unused_object():
         assert '2 objects' in magic_objects, 'Unexpected log message: %s' % magic_objects
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_network_access():
     x = Counter(name='counter')
     net = Network(x)
@@ -681,22 +709,27 @@ def test_network_access():
 
     # accessing objects
     assert net['counter'] is x
-    assert_raises(TypeError, lambda: net[123])
-    assert_raises(TypeError, lambda: net[1:3])
-    assert_raises(KeyError, lambda: net['non-existing'])
+    with pytest.raises(TypeError):
+        net[123]
+    with pytest.raises(TypeError):
+        net[1:3]
+    with pytest.raises(KeyError):
+        net['non-existing']
 
     objects = [obj for obj in net]
     assert set(objects) == set(net.objects)
 
     # deleting objects
     del net['counter']
-    assert_raises(TypeError, lambda: net.__delitem__(123))
-    assert_raises(TypeError, lambda: net.__delitem__(slice(1, 3)))
-    assert_raises(KeyError, lambda: net.__delitem__('counter'))
+    with pytest.raises(TypeError):
+        net.__delitem__(123)
+    with pytest.raises(TypeError):
+        net.__delitem__(slice(1, 3))
+    with pytest.raises(KeyError):
+        net.__delitem__('counter')
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_dependency_check():
     def create_net():
         G = NeuronGroup(10, 'v: 1', threshold='False')
@@ -711,7 +744,8 @@ def test_dependency_check():
     dependent_objects = create_net()
     # Trying to simulate the monitors/synapses without the group should fail
     for obj in dependent_objects:
-        assert_raises(ValueError, lambda: Network(obj).run(0*ms))
+        with pytest.raises(ValueError):
+            Network(obj).run(0*ms)
 
     # simulation with a magic network should work when we have an explicit
     # reference to one of the objects, but the object should be inactive and
@@ -726,7 +760,6 @@ def test_dependency_check():
         assert not obj.active
 
 
-@with_setup(teardown=restore_initial_state)
 def test_loop():
     '''
     Somewhat realistic test with a loop of magic networks
@@ -759,7 +792,7 @@ def test_loop():
         assert '4 objects' in magic_objects
 
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_magic_collect():
     '''
     Make sure all expected objects are collected in a magic network
@@ -777,12 +810,16 @@ def test_magic_collect():
     assert len(objects) == 6, ('expected %d objects, got %d' % (6, len(objects)))
 
 from contextlib import contextmanager
-from StringIO import StringIO
+from io import StringIO, BytesIO
 import sys
 
 @contextmanager
 def captured_output():
-    new_out, new_err = StringIO(), StringIO()
+    if sys.version_info[0] == 2:
+        # in Python 2, stdout/stderr works with bytes and not with unicode
+        new_out, new_err = BytesIO(), BytesIO()
+    else:
+        new_out, new_err = StringIO(), StringIO()
     old_out, old_err = sys.stdout, sys.stderr
     try:
         sys.stdout, sys.stderr = new_out, new_err
@@ -791,7 +828,7 @@ def captured_output():
         sys.stdout, sys.stderr = old_out, old_err
 
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_progress_report():
     '''
     Very basic test of progress reporting
@@ -845,19 +882,21 @@ def test_progress_report():
     assert len(calls) >= 2 and calls[0][1] == 0.0 and calls[-1][1] == 1.0
 
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_progress_report_incorrect():
     '''
     Test wrong use of the report option
     '''
     G = NeuronGroup(1, '')
     net = Network(G)
-    assert_raises(ValueError, lambda: net.run(1*ms, report='unknown'))
-    assert_raises(TypeError, lambda: net.run(1*ms, report=object()))
+    with pytest.raises(ValueError):
+        net.run(1*ms, report='unknown')
+    with pytest.raises(TypeError):
+        net.run(1*ms, report=object())
 
 
-@attr('standalone-compatible', 'multiple-runs')
-@with_setup(teardown=reinit_devices)
+@pytest.mark.standalone_compatible
+@pytest.mark.multiple_runs
 def test_multiple_runs_report_standalone():
     group = NeuronGroup(1, 'dv/dt = 1*Hz : 1')
     run(1*ms, report='text')
@@ -865,8 +904,8 @@ def test_multiple_runs_report_standalone():
     device.build(direct_call=False, **device.build_options)
 
 
-@attr('standalone-compatible', 'multiple-runs')
-@with_setup(teardown=reinit_devices)
+@pytest.mark.standalone_compatible
+@pytest.mark.multiple_runs
 def test_multiple_runs_report_standalone_2():
     group = NeuronGroup(1, 'dv/dt = 1*Hz : 1')
     run(1*ms)
@@ -874,8 +913,8 @@ def test_multiple_runs_report_standalone_2():
     device.build(direct_call=False, **device.build_options)
 
 
-@attr('standalone-compatible', 'multiple-runs')
-@with_setup(teardown=reinit_devices)
+@pytest.mark.standalone_compatible
+@pytest.mark.multiple_runs
 def test_multiple_runs_report_standalone_3():
     group = NeuronGroup(1, 'dv/dt = 1*Hz : 1')
     run(1*ms, report='text')
@@ -885,16 +924,17 @@ def test_multiple_runs_report_standalone_3():
 
 # This tests a specific limitation of the C++ standalone mode (cannot mix
 # multiple report methods)
-@attr('cpp_standalone', 'standalone-only')
-@with_setup(teardown=reinit_devices)
+@pytest.mark.cpp_standalone
+@pytest.mark.standalone_only
 def test_multiple_runs_report_standalone_incorrect():
     set_device('cpp_standalone', build_on_run=False)
     group = NeuronGroup(1, 'dv/dt = 1*Hz : 1')
     run(1*ms, report='text')
-    assert_raises(NotImplementedError, lambda: run(1*ms, report='stderr'))
+    with pytest.raises(NotImplementedError):
+        run(1*ms, report='stderr')
 
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_store_restore():
     source = NeuronGroup(10, '''dv/dt = rates : 1
                                 rates : Hz''', threshold='v>1', reset='v=0')
@@ -936,7 +976,7 @@ def test_store_restore():
     assert net.t == 10 * ms
 
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_store_restore_to_file():
     filename = tempfile.mktemp(suffix='state', prefix='brian_test')
     source = NeuronGroup(10, '''dv/dt = rates : 1
@@ -978,7 +1018,7 @@ def test_store_restore_to_file():
     except OSError:
         pass
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_store_restore_to_file_new_objects():
     # A more realistic test where the objects are completely re-created
     filename = tempfile.mktemp(suffix='state', prefix='brian_test')
@@ -1032,7 +1072,7 @@ def test_store_restore_to_file_new_objects():
         pass
 
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_store_restore_to_file_differing_nets():
     # Check that the store/restore mechanism is not used with differing
     # networks
@@ -1048,13 +1088,14 @@ def test_store_restore_to_file_differing_nets():
                                    name='source_2')
     mon = SpikeMonitor(source_2, name='monitor')
     net = Network(source_2, mon)
-    assert_raises(KeyError, lambda: net.restore(filename=filename))
+    with pytest.raises(KeyError):
+        net.restore(filename=filename)
 
     net = Network(source)  # Without the monitor
-    assert_raises(KeyError, lambda: net.restore(filename=filename))
+    with pytest.raises(KeyError):
+        net.restore(filename=filename)
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_store_restore_magic():
     source = NeuronGroup(10, '''dv/dt = rates : 1
                                 rates : Hz''', threshold='v>1', reset='v=0')
@@ -1091,8 +1132,7 @@ def test_store_restore_magic():
     assert_equal(spike_times, spike_mon.t_[:])
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_store_restore_magic_to_file():
     filename = tempfile.mktemp(suffix='state', prefix='brian_test')
     source = NeuronGroup(10, '''dv/dt = rates : 1
@@ -1134,8 +1174,7 @@ def test_store_restore_magic_to_file():
         pass
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_store_restore_spikequeue():
     # See github issue #938
     source = SpikeGeneratorGroup(1, [0], [0] * ms)
@@ -1153,9 +1192,28 @@ def test_store_restore_spikequeue():
     run(2 * defaultclock.dt)
     assert target.v[0] == 1
 
+@pytest.mark.skipif(not isinstance(get_device(), RuntimeDevice),
+                    reason='Getting/setting random number state only supported '
+                           'for runtime device.')
+def test_restore_with_random_state():
+    group = NeuronGroup(10, 'dv/dt = -v/(10*ms) + (10*ms)**-0.5*xi : 1',
+                        method='euler')
+    group.v = 'rand()'
+    mon = StateMonitor(group, 'v', record=True)
+    store()
+    run(10*ms)
+    old_v = np.array(group.v)
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+    restore()  # Random state is not restored
+    run(10 * ms)
+    assert np.var(old_v - group.v) > 0  # very basic test for difference
+
+    restore(restore_random_state=True)  # Random state is restored
+    run(10 * ms)
+    assert_equal(old_v, group.v)
+
+
+@pytest.mark.codegen_independent
 def test_defaultclock_dt_changes():
     BrianLogger.suppress_name('resolution_conflict')
     for dt in [0.1*ms, 0.01*ms, 0.5*ms, 1*ms, 3.3*ms]:
@@ -1166,8 +1224,8 @@ def test_defaultclock_dt_changes():
         net.run(2*dt)
         assert_equal(mon.t[:], [0, dt/ms]*ms)
 
-@attr('standalone-compatible', 'multiple-runs')
-@with_setup(teardown=reinit_devices)
+@pytest.mark.standalone_compatible
+@pytest.mark.multiple_runs
 def test_dt_changes_between_runs():
     defaultclock.dt = 0.1*ms
     G = NeuronGroup(1, 'v:1')
@@ -1182,8 +1240,7 @@ def test_dt_changes_between_runs():
     assert_allclose(mon.t[:],
                     [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 1., 1.1, 1.2, 1.3, 1.4]*ms)
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_dt_restore():
     defaultclock.dt = 0.5*ms
     G = NeuronGroup(1, 'dv/dt = -v/(10*ms) : 1')
@@ -1203,8 +1260,7 @@ def test_dt_restore():
     assert_equal(mon.t[:], [0, 0.5]*ms)
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_continuation():
     defaultclock.dt = 1*ms
     G = NeuronGroup(1, 'dv/dt = -v / (10*ms) : 1')
@@ -1225,8 +1281,7 @@ def test_continuation():
     assert_equal(mon.v[:], mon2.v[:])
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_get_set_states():
     G = NeuronGroup(10, 'v:1', name='a_neurongroup')
     G.v = 'i'
@@ -1235,7 +1290,9 @@ def test_get_set_states():
     states2 = magic_network.get_states()
     states3 = net.get_states(read_only_variables=False)
     assert set(states1.keys()) == set(states2.keys()) == set(states3.keys()) == {'a_neurongroup'}
-    assert set(states1['a_neurongroup'].keys()) == set(states2['a_neurongroup'].keys()) == {'i', 'dt', 'N', 't', 'v'}
+    assert (set(states1['a_neurongroup'].keys()) ==
+            set(states2['a_neurongroup'].keys()) ==
+            {'i', 'dt', 'N', 't', 'v', 't_in_timesteps'})
     assert set(states3['a_neurongroup']) == {'v'}
 
     # Try re-setting the state
@@ -1244,8 +1301,7 @@ def test_get_set_states():
     assert_equal(G.v, np.arange(10))
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_multiple_runs_defaultclock():
     defaultclock.dt = 0.1*ms
     G = NeuronGroup(1, 'dv/dt = -v / (10*ms) : 1')
@@ -1260,8 +1316,7 @@ def test_multiple_runs_defaultclock():
     net.run(1*ms)
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_multiple_runs_defaultclock_incorrect():
     defaultclock.dt = 0.1*ms
     G = NeuronGroup(1, 'dv/dt = -v / (10*ms) : 1')
@@ -1271,11 +1326,11 @@ def test_multiple_runs_defaultclock_incorrect():
     # The new dt is not compatible with the previous time since we cannot
     # continue at 0.5ms with a dt of 1ms
     defaultclock.dt = 1*ms
-    assert_raises(ValueError, lambda: net.run(1*ms))
+    with pytest.raises(ValueError):
+        net.run(1*ms)
 
 
-@attr('standalone-compatible')
-@with_setup(teardown=reinit_devices)
+@pytest.mark.standalone_compatible
 def test_profile():
     G = NeuronGroup(10, 'dv/dt = -v / (10*ms) : 1', threshold='v>1',
                     reset='v=0', name='profile_test')
@@ -1297,17 +1352,17 @@ def test_profile():
     assert all([t>=0*second for _, t in info])
 
 
-@attr('standalone-compatible')
-@with_setup(teardown=reinit_devices)
+@pytest.mark.standalone_compatible
 def test_profile_off():
     G = NeuronGroup(10, 'dv/dt = -v / (10*ms) : 1', threshold='v>1',
                     reset='v=0', name='profile_test')
     net = Network(G)
     net.run(1*ms, profile=False)
-    assert_raises(ValueError, lambda: profiling_summary(net))
+    with pytest.raises(ValueError):
+        profiling_summary(net)
 
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_profile_ipython_html():
     G = NeuronGroup(10, 'dv/dt = -v / (10*ms) : 1', threshold='v>1',
                     reset='v=0', name='profile_test')
@@ -1318,8 +1373,7 @@ def test_profile_ipython_html():
     assert len(summary._repr_html_())
 
 
-@attr('codegen-independent')
-@with_setup(teardown=restore_initial_state)
+@pytest.mark.codegen_independent
 def test_magic_scope():
     '''
     Check that `start_scope` works as expected.
@@ -1335,8 +1389,7 @@ def test_magic_scope():
     assert objs2=={'G3', 'G4'}
 
 
-@attr('standalone-compatible')
-@with_setup(teardown=reinit_devices)
+@pytest.mark.standalone_compatible
 def test_runtime_rounding():
     # Test that runtime and standalone round in the same way, see github issue
     # #695 for details
@@ -1346,7 +1399,7 @@ def test_runtime_rounding():
     run(defaultclock.dt * 250)
     assert len(mon.t) == 250
 
-@attr('codegen-independent')
+@pytest.mark.codegen_independent
 def test_small_runs():
     # One long run and multiple small runs should give the same results
     group_1 = NeuronGroup(10, 'dv/dt = -v / (10*ms) : 1')
@@ -1370,9 +1423,29 @@ def test_small_runs():
     assert_allclose(mon_1.t_[:], mon_2.t_[:])
     assert_allclose(mon_1.v_[:], mon_2.v_[:])
 
+@pytest.mark.codegen_independent
+def test_both_equal():
+    #check all objects added by Network.add() also have their contained_objects added to 'Network'
+    tau = 10*ms
+    diff_eqn='''dv/dt = (1-v)/tau : 1'''
+    chg_code='''v = 2*v'''
 
-@attr('codegen-independent')
-@with_setup(teardown=reinit_devices)
+    Ng = NeuronGroup(1,diff_eqn,method='exact')
+    M1 = StateMonitor(Ng, 'v', record=True)
+    netObj = Network(Ng,M1)
+    Ng.run_regularly(chg_code, dt=20*ms)
+    netObj.run(100*ms)
+
+    start_scope()
+    Ng = NeuronGroup(1,diff_eqn,method='exact')
+    M2 = StateMonitor(Ng, 'v', record=True)
+    Ng.run_regularly(chg_code, dt=20*ms)
+    run(100*ms)
+
+    assert (M1.v == M2.v).all()
+
+
+@pytest.mark.codegen_independent
 def test_long_run_dt_change():
     # Check that the dt check is not too restrictive, see issue #730 for details
     group = NeuronGroup(1, '')  # does nothing...
@@ -1382,8 +1455,8 @@ def test_long_run_dt_change():
     defaultclock.dt = 0.01*ms
     run(1*second)
 
-@attr('standalone-compatible', 'multiple-runs')
-@with_setup(teardown=reinit_devices)
+@pytest.mark.standalone_compatible
+@pytest.mark.multiple_runs
 def test_multiple_runs_constant_change():
     const_v = 1
     group = NeuronGroup(1, 'v = const_v : 1')
@@ -1395,8 +1468,8 @@ def test_multiple_runs_constant_change():
     assert_equal(mon.v[0], [1, 2])
 
 
-@attr('standalone-compatible', 'multiple-runs')
-@with_setup(teardown=reinit_devices)
+@pytest.mark.standalone_compatible
+@pytest.mark.multiple_runs
 def test_multiple_runs_function_change():
     inp = TimedArray([1, 2], dt=defaultclock.dt)
     group = NeuronGroup(1, 'v = inp(t) : 1')
@@ -1430,6 +1503,7 @@ if __name__ == '__main__':
             test_network_stop,
             test_network_operations,
             test_incorrect_network_operations,
+            test_network_operations_name,
             test_network_active_flag,
             test_network_t,
             test_incorrect_dt_defaultclock,
@@ -1468,10 +1542,11 @@ if __name__ == '__main__':
             test_magic_scope,
             test_runtime_rounding,
             test_small_runs,
+            test_both_equal,
             test_long_run_dt_change,
             test_multiple_runs_constant_change,
             test_multiple_runs_function_change
             ]:
         set_device(all_devices['runtime'])
         t()
-        reinit_devices()
+        reinit_and_delete()
